use axum::{
    extract::{Path, Query},
    response::{Html, IntoResponse},
    Extension,
};
use axum_extra::extract::CookieJar;
use tetratto_core::model::{permissions::FinePermission, stacks::StackPrivacy, Error, auth::User};
use crate::{assets::initial_context, get_lang, get_user_from_token, State};
use super::{render_error, PaginatedQuery};

/// `/stacks`
pub async fn list_request(jar: CookieJar, Extension(data): Extension<State>) -> impl IntoResponse {
    let data = data.read().await;
    let user = match get_user_from_token!(jar, data.0) {
        Some(ua) => ua,
        None => {
            return Err(Html(
                render_error(Error::NotAllowed, &jar, &data, &None).await,
            ));
        }
    };

    let list = match data.0.get_stacks_by_owner(user.id).await {
        Ok(p) => p,
        Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
    };

    let lang = get_lang!(jar, data.0);
    let mut context = initial_context(&data.0.0, lang, &Some(user)).await;

    context.insert("list", &list);

    // return
    Ok(Html(data.1.render("stacks/list.html", &context).unwrap()))
}

/// `/stacks/{id}`
pub async fn posts_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
    Query(req): Query<PaginatedQuery>,
) -> impl IntoResponse {
    let data = data.read().await;
    let user = match get_user_from_token!(jar, data.0) {
        Some(ua) => ua,
        None => {
            return Err(Html(
                render_error(Error::NotAllowed, &jar, &data, &None).await,
            ));
        }
    };

    let stack = match data.0.get_stack_by_id(id).await {
        Ok(s) => s,
        Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
    };

    if stack.privacy == StackPrivacy::Private
        && user.id != stack.owner
        && !user.permissions.check(FinePermission::MANAGE_STACKS)
    {
        return Err(Html(
            render_error(Error::NotAllowed, &jar, &data, &None).await,
        ));
    }

    let ignore_users = data.0.get_userblocks_receivers(user.id).await;
    let list = match data
        .0
        .get_stack_posts(user.id, stack.id, 12, req.page, &ignore_users)
        .await
    {
        Ok(l) => l,
        Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
    };

    let lang = get_lang!(jar, data.0);
    let mut context = initial_context(&data.0.0, lang, &Some(user)).await;

    context.insert("page", &req.page);
    context.insert("stack", &stack);
    context.insert("list", &list);

    // return
    Ok(Html(data.1.render("stacks/posts.html", &context).unwrap()))
}

/// `/stacks/{id}/manage`
pub async fn manage_request(
    jar: CookieJar,
    Extension(data): Extension<State>,
    Path(id): Path<usize>,
) -> impl IntoResponse {
    let data = data.read().await;
    let user = match get_user_from_token!(jar, data.0) {
        Some(ua) => ua,
        None => {
            return Err(Html(
                render_error(Error::NotAllowed, &jar, &data, &None).await,
            ));
        }
    };

    let stack = match data.0.get_stack_by_id(id).await {
        Ok(s) => s,
        Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
    };

    if user.id != stack.owner && !user.permissions.check(FinePermission::MANAGE_STACKS) {
        return Err(Html(
            render_error(Error::NotAllowed, &jar, &data, &None).await,
        ));
    }

    let mut users: Vec<User> = Vec::new();

    for uid in &stack.users {
        users.push(match data.0.get_user_by_id(uid.to_owned()).await {
            Ok(ua) => ua,
            Err(e) => return Err(Html(render_error(e, &jar, &data, &Some(user)).await)),
        });
    }

    let lang = get_lang!(jar, data.0);
    let mut context = initial_context(&data.0.0, lang, &Some(user)).await;

    context.insert("stack", &stack);
    context.insert("users", &users);

    // return
    Ok(Html(data.1.render("stacks/manage.html", &context).unwrap()))
}
